@@ -1,4 +1,4 @@
-"""DiffSync core classes and logic.
+"""DiffSync front-end classes and logic.
 
 Copyright (c) 2020 Network To Code, LLC <info@networktocode.com>
 
@@ -15,73 +15,16 @@ See the License for the specific language governing permissions and
 limitations under the License.
 """
 from collections import defaultdict
-from collections.abc import Iterable as ABCIterable, Mapping as ABCMapping
-import enum
 from inspect import isclass
-from typing import ClassVar, Dict, Iterable, List, Mapping, MutableMapping, Optional, Text, Tuple, Type, Union
+from typing import ClassVar, Dict, List, Mapping, MutableMapping, Optional, Text, Tuple, Type, Union
 
-from pydantic import BaseModel
+from pydantic import BaseModel, PrivateAttr
 import structlog  # type: ignore
 
-from .diff import Diff, DiffElement
-from .utils import intersection
-from .exceptions import (
-    ObjectNotCreated,
-    ObjectNotUpdated,
-    ObjectNotDeleted,
-    ObjectCrudException,
-    ObjectAlreadyExists,
-    ObjectStoreWrongType,
-    ObjectNotFound,
-)
-
-
-class DiffSyncModelFlags(enum.Flag):
-    """Flags that can be set on a DiffSyncModel class or instance to affect its usage."""
-
-    NONE = 0
-
-    IGNORE = 0b1
-    """Do not render diffs containing this model; do not make any changes to this model when synchronizing.
-
-    Can be used to indicate a model instance that exists but should not be changed by DiffSync.
-    """
-
-    SKIP_CHILDREN_ON_DELETE = 0b10
-    """When deleting this model, do not recursively delete its children.
-
-    Can be used for the case where deletion of a model results in the automatic deletion of all its children.
-    """
-
-
-class DiffSyncFlags(enum.Flag):
-    """Flags that can be passed to a sync_* or diff_* call to affect its behavior."""
-
-    NONE = 0
-
-    CONTINUE_ON_FAILURE = 0b1
-    """Continue synchronizing even if failures are encountered when syncing individual models."""
-
-    SKIP_UNMATCHED_SRC = 0b10
-    """Ignore objects that only exist in the source/"from" DiffSync when determining diffs and syncing.
-
-    If this flag is set, no new objects will be created in the target/"to" DiffSync.
-    """
-
-    SKIP_UNMATCHED_DST = 0b100
-    """Ignore objects that only exist in the target/"to" DiffSync when determining diffs and syncing.
-
-    If this flag is set, no objects will be deleted from the target/"to" DiffSync.
-    """
-
-    SKIP_UNMATCHED_BOTH = SKIP_UNMATCHED_SRC | SKIP_UNMATCHED_DST
-
-    LOG_UNCHANGED_RECORDS = 0b1000
-    """If this flag is set, a log message will be generated during synchronization for each model, even unchanged ones.
-
-    By default, when this flag is unset, only models that have actual changes to synchronize will be logged.
-    This flag is off by default to reduce the default verbosity of DiffSync, but can be enabled when debugging.
-    """
+from .diff import Diff
+from .enum import DiffSyncModelFlags, DiffSyncFlags, DiffSyncStatus
+from .exceptions import ObjectAlreadyExists, ObjectStoreWrongType, ObjectNotFound
+from .helpers import DiffSyncDiffer, DiffSyncSyncer
 
 
 class DiffSyncModel(BaseModel):
@@ -144,6 +87,12 @@ class DiffSyncModel(BaseModel):
     diffsync: Optional["DiffSync"] = None
     """Optional: the DiffSync instance that owns this model instance."""
 
+    _status: DiffSyncStatus = PrivateAttr(DiffSyncStatus.SUCCESS)
+    """Status of the last attempt at creating/updating/deleting this model."""
+
+    _status_message: str = PrivateAttr("")
+    """Message, if any, associated with the create/update/delete status value."""
+
     class Config:  # pylint: disable=too-few-public-methods
         """Pydantic class configuration."""
 
@@ -214,17 +163,25 @@ class DiffSyncModel(BaseModel):
                 output += f": {child_ids}"
             else:
                 for child_id in child_ids:
-                    child = self.diffsync.get(modelname, child_id)
-                    if not child:
-                        output += f"\n{margin}    {child_id} (details unavailable)"
-                    else:
+                    try:
+                        child = self.diffsync.get(modelname, child_id)
                         output += "\n" + child.str(include_children=include_children, indent=indent + 4)
+                    except ObjectNotFound:
+                        output += f"\n{margin}    {child_id} (ERROR: details unavailable)"
         return output
 
+    def set_status(self, status: DiffSyncStatus, message: Text = ""):
+        """Update the status (and optionally status message) of this model in response to a create/update/delete call."""
+        self._status = status
+        self._status_message = message
+
     @classmethod
     def create(cls, diffsync: "DiffSync", ids: Mapping, attrs: Mapping) -> Optional["DiffSyncModel"]:
         """Instantiate this class, along with any platform-specific data creation.
 
+        Subclasses must call `super().create()`; they may wish to then override the default status information
+        by calling `set_status()` to provide more context (such as details of any interactions with underlying systems).
+
         Args:
             diffsync: The master data store for other DiffSyncModel instances that we might need to reference
             ids: Dictionary of unique-identifiers needed to create the new object
@@ -237,11 +194,16 @@ class DiffSyncModel(BaseModel):
         Raises:
             ObjectNotCreated: if an error occurred.
         """
-        return cls(**ids, diffsync=diffsync, **attrs)
+        model = cls(**ids, diffsync=diffsync, **attrs)
+        model.set_status(DiffSyncStatus.SUCCESS, "Created successfully")
+        return model
 
     def update(self, attrs: Mapping) -> Optional["DiffSyncModel"]:
         """Update the attributes of this instance, along with any platform-specific data updates.
 
+        Subclasses must call `super().update()`; they may wish to then override the default status information
+        by calling `set_status()` to provide more context (such as details of any interactions with underlying systems).
+
         Args:
             attrs: Dictionary of attributes to update on the object
 
@@ -255,11 +217,16 @@ class DiffSyncModel(BaseModel):
         for attr, value in attrs.items():
             # TODO: enforce that only attrs in self._attributes can be updated in this way?
             setattr(self, attr, value)
+
+        self.set_status(DiffSyncStatus.SUCCESS, "Updated successfully")
         return self
 
     def delete(self) -> Optional["DiffSyncModel"]:
         """Delete any platform-specific data corresponding to this instance.
 
+        Subclasses must call `super().delete()`; they may wish to then override the default status information
+        by calling `set_status()` to provide more context (such as details of any interactions with underlying systems).
+
         Returns:
             DiffSyncModel: this instance, if all data was successfully deleted.
             None: if data deletion failed in such a way that child objects of this model should not be deleted.
@@ -267,6 +234,7 @@ class DiffSyncModel(BaseModel):
         Raises:
             ObjectNotDeleted: if an error occurred.
         """
+        self.set_status(DiffSyncStatus.SUCCESS, "Deleted successfully")
         return self
 
     @classmethod
@@ -336,6 +304,10 @@ class DiffSyncModel(BaseModel):
             return "__".join([str(getattr(self, key)) for key in self._shortname])
         return self.get_unique_id()
 
+    def get_status(self) -> Tuple[DiffSyncStatus, Text]:
+        """Get the status of the last create/update/delete operation on this object, and any associated message."""
+        return (self._status, self._status_message)
+
     def add_child(self, child: "DiffSyncModel"):
         """Add a child reference to an object.
 
@@ -350,7 +322,8 @@ class DiffSyncModel(BaseModel):
 
         if child_type not in self._children:
             raise ObjectStoreWrongType(
-                f"Unable to store {child_type} as a child; valid types are {sorted(self._children.keys())}"
+                f"Unable to store {child_type} as a child of {self.get_type()}; "
+                f"valid types are {sorted(self._children.keys())}"
             )
 
         attr_name = self._children[child_type]
@@ -372,7 +345,8 @@ class DiffSyncModel(BaseModel):
 
         if child_type not in self._children:
             raise ObjectStoreWrongType(
-                f"Unable to store {child_type} as a child; valid types are {sorted(self._children.keys())}"
+                f"Unable to find and delete {child_type} as a child of {self.get_type()}; "
+                f"valid types are {sorted(self._children.keys())}"
             )
 
         attr_name = self._children[child_type]
@@ -428,6 +402,13 @@ class DiffSync:
                     f'Incorrect field name - {value.__name__} has type name "{value.get_type()}", not "{name}"'
                 )
 
+        for name in cls.top_level:
+            if not hasattr(cls, name):
+                raise AttributeError(f'top_level references attribute "{name}" but it is not a class attribute!')
+            value = getattr(cls, name)
+            if not isclass(value) or not issubclass(value, DiffSyncModel):
+                raise AttributeError(f'top_level references attribute "{name}" but it is not a DiffSyncModel subclass!')
+
     def __str__(self):
         """String representation of a DiffSync."""
         if self.type != self.name:
@@ -455,6 +436,8 @@ class DiffSync:
         margin = " " * indent
         output = ""
         for modelname in self.top_level:
+            if output:
+                output += "\n"
             output += f"{margin}{modelname}"
             models = self.get_all(modelname)
             if not models:
@@ -476,13 +459,11 @@ class DiffSync:
             diff_class (class): Diff or subclass thereof to use to calculate the diffs to use for synchronization
             flags (DiffSyncFlags): Flags influencing the behavior of this sync.
         """
-        log = self._log.bind(src=source, dst=self, flags=flags).unbind("diffsync")
         diff = self.diff_from(source, diff_class=diff_class, flags=flags)
-
-        log.info("Beginning sync")
-        for child in diff.get_children():
-            self._sync_from_diff_element(child, flags=flags, logger=log)
-        log.info("Sync complete")
+        syncer = DiffSyncSyncer(diff=diff, src_diffsync=source, dst_diffsync=self, flags=flags)
+        result = syncer.perform_sync()
+        if result:
+            self.sync_complete(source, diff, flags, syncer.base_logger)
 
     def sync_to(self, target: "DiffSync", diff_class: Type[Diff] = Diff, flags: DiffSyncFlags = DiffSyncFlags.NONE):
         """Synchronize data from the current DiffSync object into the given target DiffSync object.
@@ -494,87 +475,27 @@ class DiffSync:
         """
         target.sync_from(self, diff_class=diff_class, flags=flags)
 
-    def _sync_from_diff_element(
+    def sync_complete(
         self,
-        element: DiffElement,
+        source: "DiffSync",
+        diff: Diff,
         flags: DiffSyncFlags = DiffSyncFlags.NONE,
         logger: structlog.BoundLogger = None,
-        parent_model: DiffSyncModel = None,
     ):
-        """Synchronize a given DiffElement (and its children, if any) into this DiffSync.
+        """Callback triggered after a `sync_from` operation has completed and updated the model data of this instance.
 
-        Helper method for `sync_from`/`sync_to`; this generally shouldn't be called on its own.
+        Note that this callback is **only** triggered if the sync actually resulted in data changes. If there are no
+        detected changes, this callback will **not** be called.
+
+        The default implementation does nothing, but a subclass could use this, for example, to perform bulk updates
+        to a backend (such as a file) that doesn't readily support incremental updates to individual records.
 
         Args:
-            element: DiffElement to synchronize diffs from
-            flags (DiffSyncFlags): Flags influencing the behavior of this sync.
-            logger: Parent logging context
-            parent_model: Parent object to update (`add_child`/`remove_child`) if the sync creates/deletes an object.
+          source: The DiffSync whose data was used to update this instance.
+          diff: The Diff calculated prior to the sync operation.
+          flags: Any flags that influenced the sync.
+          logger: Logging context for the sync.
         """
-        # pylint: disable=too-many-branches
-        # GFM: I made a few attempts at refactoring this to reduce the branching, but found that it was less readable.
-        # So let's live with the slightly too high number of branches (14/12) for now.
-        log = logger or self._log
-        object_class = getattr(self, element.type)
-        obj = self.get(object_class, element.keys)
-        # Get the attributes that actually differ between source and dest
-        diffs = element.get_attrs_diffs()
-        log = log.bind(
-            action=element.action,
-            model=object_class.get_type(),
-            unique_id=object_class.create_unique_id(**element.keys),
-            diffs=diffs,
-        )
-
-        try:
-            if element.action == "create":
-                log.debug("Attempting object creation")
-                if obj:
-                    raise ObjectNotCreated(f"Failed to create {object_class.get_type()} {element.keys} - it exists!")
-                obj = object_class.create(diffsync=self, ids=element.keys, attrs=diffs["src"])
-                log.info("Created successfully", status="success")
-            elif element.action == "update":
-                log.debug("Attempting object update")
-                if not obj:
-                    raise ObjectNotUpdated(f"Failed to update {object_class.get_type()} {element.keys} - not found!")
-                obj = obj.update(attrs=diffs["src"])
-                log.info("Updated successfully", status="success")
-            elif element.action == "delete":
-                log.debug("Attempting object deletion")
-                if not obj:
-                    raise ObjectNotDeleted(f"Failed to delete {object_class.get_type()} {element.keys} - not found!")
-                obj = obj.delete()
-                log.info("Deleted successfully", status="success")
-            else:
-                if flags & DiffSyncFlags.LOG_UNCHANGED_RECORDS:
-                    log.debug("No action needed", status="success")
-        except ObjectCrudException as exception:
-            log.error(str(exception), status="error")
-            if not flags & DiffSyncFlags.CONTINUE_ON_FAILURE:
-                raise
-        else:
-            if obj is None:
-                log.warning("Non-fatal failure encountered", status="failure")
-
-        if obj is None:
-            log.warning("Not syncing children")
-            return
-
-        if element.action == "create":
-            if parent_model:
-                parent_model.add_child(obj)
-            self.add(obj)
-        elif element.action == "delete":
-            if parent_model:
-                parent_model.remove_child(obj)
-            if obj.model_flags & DiffSyncModelFlags.SKIP_CHILDREN_ON_DELETE:
-                # We don't need to process the child objects, but we do need to discard them
-                self.remove(obj, remove_children=True)
-                return
-            self.remove(obj)
-
-        for child in element.get_children():
-            self._sync_from_diff_element(child, flags=flags, parent_model=obj, logger=logger)
 
     # ------------------------------------------------------------------------------
     # Diff calculation and construction
@@ -611,12 +532,16 @@ class DiffSync:
 
     def get(
         self, obj: Union[Text, DiffSyncModel, Type[DiffSyncModel]], identifier: Union[Text, Mapping]
-    ) -> Optional[DiffSyncModel]:
+    ) -> DiffSyncModel:
         """Get one object from the data store based on its unique id.
 
         Args:
             obj: DiffSyncModel class or instance, or modelname string, that defines the type of the object to retrieve
             identifier: Unique ID of the object to retrieve, or dict of unique identifier keys/values
+
+        Raises:
+            ValueError: if obj is a str and identifier is a dict (can't convert dict into a uid str without a model class)
+            ObjectNotFound: if the requested object is not present
         """
         if isinstance(obj, str):
             modelname = obj
@@ -633,13 +558,14 @@ class DiffSync:
         elif object_class:
             uid = object_class.create_unique_id(**identifier)
         else:
-            self._log.warning(
-                f"Tried to look up a {modelname} by identifier {identifier}, "
-                "but don't know how to convert that to a uid string",
+            raise ValueError(
+                f"Invalid args: ({obj}, {identifier}): "
+                f"either {obj} should be a class/instance or {identifier} should be a str"
             )
-            return None
 
-        return self._data[modelname].get(uid)
+        if uid not in self._data[modelname]:
+            raise ObjectNotFound(f"{modelname} {uid} not present in {self.name}")
+        return self._data[modelname][uid]
 
     def get_all(self, obj: Union[Text, DiffSyncModel, Type[DiffSyncModel]]):
         """Get all objects of a given type.
@@ -665,17 +591,20 @@ class DiffSync:
         Args:
             uids: List of unique id / key identifying object in the database.
             obj: DiffSyncModel class or instance, or modelname string, that defines the type of the objects to retrieve
+
+        Raises:
+            ObjectNotFound: if any of the requested UIDs are not found in the store
         """
         if isinstance(obj, str):
             modelname = obj
         else:
             modelname = obj.get_type()
 
-        # TODO: should this raise an exception if any or all of the uids are not found?
         results = []
         for uid in uids:
-            if uid in self._data[modelname]:
-                results.append(self._data[modelname][uid])
+            if uid not in self._data[modelname]:
+                raise ObjectNotFound(f"{modelname} {uid} not present in {self.name}")
+            results.append(self._data[modelname][uid])
         return results
 
     def add(self, obj: DiffSyncModel):
@@ -712,7 +641,7 @@ class DiffSync:
         uid = obj.get_unique_id()
 
         if uid not in self._data[modelname]:
-            raise ObjectNotFound(f"Object {uid} not present")
+            raise ObjectNotFound(f"{modelname} {uid} not present in {self.name}")
 
         if obj.diffsync is self:
             obj.diffsync = None
@@ -722,199 +651,13 @@ class DiffSync:
         if remove_children:
             for child_type, child_fieldname in obj.get_children_mapping().items():
                 for child_id in getattr(obj, child_fieldname):
-                    child_obj = self.get(child_type, child_id)
-                    if child_obj:
+                    try:
+                        child_obj = self.get(child_type, child_id)
                         self.remove(child_obj, remove_children=remove_children)
+                    except ObjectNotFound:
+                        # Since this is "cleanup" code, log an error and continue, instead of letting the exception raise
+                        self._log.error(f"Unable to remove child {child_id} of {modelname} {uid} - not found!")
 
 
 # DiffSyncModel references DiffSync and DiffSync references DiffSyncModel. Break the typing loop:
 DiffSyncModel.update_forward_refs()
-
-
-class DiffSyncDiffer:
-    """Helper class implementing diff calculation logic for DiffSync.
-
-    Independent from Diff and DiffElement as those classes are purely data objects, while this stores some state.
-    """
-
-    def __init__(
-        self, src_diffsync: DiffSync, dst_diffsync: DiffSync, flags: DiffSyncFlags, diff_class: Type[Diff] = Diff
-    ):
-        """Create a DiffSyncDiffer for calculating diffs between the provided DiffSync instances."""
-        self.src_diffsync = src_diffsync
-        self.dst_diffsync = dst_diffsync
-        self.flags = flags
-
-        self.logger = structlog.get_logger().new(src=src_diffsync, dst=dst_diffsync, flags=flags)
-        self.diff_class = diff_class
-        self.diff: Optional[Diff] = None
-
-    def calculate_diffs(self) -> Diff:
-        """Calculate diffs between the src and dst DiffSync objects and return the resulting Diff."""
-        if self.diff is not None:
-            return self.diff
-
-        self.logger.info("Beginning diff calculation")
-        self.diff = self.diff_class()
-        for obj_type in intersection(self.dst_diffsync.top_level, self.src_diffsync.top_level):
-            diff_elements = self.diff_object_list(
-                src=self.src_diffsync.get_all(obj_type), dst=self.dst_diffsync.get_all(obj_type),
-            )
-
-            for diff_element in diff_elements:
-                self.diff.add(diff_element)
-
-        self.logger.info("Diff calculation complete")
-        self.diff.complete()
-        return self.diff
-
-    def diff_object_list(self, src: Iterable[DiffSyncModel], dst: Iterable[DiffSyncModel]) -> List[DiffElement]:
-        """Calculate diffs between two lists of like objects.
-
-        Helper method to `calculate_diffs`, usually doesn't need to be called directly.
-
-        These helper methods work in a recursive cycle:
-        diff_object_list -> diff_object_pair -> diff_child_objects -> diff_object_list -> etc.
-        """
-        diff_elements = []
-
-        if isinstance(src, ABCIterable) and isinstance(dst, ABCIterable):
-            # Convert a list of DiffSyncModels into a dict using the unique_ids as keys
-            dict_src = {item.get_unique_id(): item for item in src} if not isinstance(src, ABCMapping) else src
-            dict_dst = {item.get_unique_id(): item for item in dst} if not isinstance(dst, ABCMapping) else dst
-
-            combined_dict = {}
-            for uid in dict_src:
-                combined_dict[uid] = (dict_src.get(uid), dict_dst.get(uid))
-            for uid in dict_dst:
-                combined_dict[uid] = (dict_src.get(uid), dict_dst.get(uid))
-        else:
-            # In the future we might support set, etc...
-            raise TypeError(f"Type combination {type(src)}/{type(dst)} is not supported... for now")
-
-        self.validate_objects_for_diff(combined_dict.values())
-
-        for uid in combined_dict:
-            src_obj, dst_obj = combined_dict[uid]
-            diff_element = self.diff_object_pair(src_obj, dst_obj)
-
-            if diff_element:
-                diff_elements.append(diff_element)
-
-        return diff_elements
-
-    @staticmethod
-    def validate_objects_for_diff(object_pairs: Iterable[Tuple[Optional[DiffSyncModel], Optional[DiffSyncModel]]]):
-        """Check whether all DiffSyncModels in the given dictionary are valid for comparison to one another.
-
-        Helper method for `diff_object_list`.
-
-        Raises:
-            TypeError: If any pair of objects in the dict have differing get_type() values.
-            ValueError: If any pair of objects in the dict have differing get_shortname() or get_identifiers() values.
-        """
-        for src_obj, dst_obj in object_pairs:
-            # TODO: should we check/enforce whether all source models have the same DiffSync, whether all dest likewise?
-            # TODO: should we check/enforce whether ALL DiffSyncModels in this dict have the same get_type() output?
-            if src_obj and dst_obj:
-                if src_obj.get_type() != dst_obj.get_type():
-                    raise TypeError(f"Type mismatch: {src_obj.get_type()} vs {dst_obj.get_type()}")
-                if src_obj.get_shortname() != dst_obj.get_shortname():
-                    raise ValueError(f"Shortname mismatch: {src_obj.get_shortname()} vs {dst_obj.get_shortname()}")
-                if src_obj.get_identifiers() != dst_obj.get_identifiers():
-                    raise ValueError(f"Keys mismatch: {src_obj.get_identifiers()} vs {dst_obj.get_identifiers()}")
-
-    def diff_object_pair(
-        self, src_obj: Optional[DiffSyncModel], dst_obj: Optional[DiffSyncModel]
-    ) -> Optional["DiffElement"]:
-        """Diff the two provided DiffSyncModel objects and return a DiffElement or None.
-
-        Helper method to `calculate_diffs`, usually doesn't need to be called directly.
-
-        These helper methods work in a recursive cycle:
-        diff_object_list -> diff_object_pair -> diff_child_objects -> diff_object_list -> etc.
-        """
-        if src_obj:
-            model = src_obj.get_type()
-            unique_id = src_obj.get_unique_id()
-            shortname = src_obj.get_shortname()
-            keys = src_obj.get_identifiers()
-        elif dst_obj:
-            model = dst_obj.get_type()
-            unique_id = dst_obj.get_unique_id()
-            shortname = dst_obj.get_shortname()
-            keys = dst_obj.get_identifiers()
-        else:
-            raise RuntimeError("diff_object_pair() called with neither src_obj nor dst_obj??")
-
-        log = self.logger.bind(model=model, unique_id=unique_id)
-        if self.flags & DiffSyncFlags.SKIP_UNMATCHED_SRC and not dst_obj:
-            log.debug("Skipping unmatched source object")
-            return None
-        if self.flags & DiffSyncFlags.SKIP_UNMATCHED_DST and not src_obj:
-            log.debug("Skipping unmatched dest object")
-            return None
-        if src_obj and src_obj.model_flags & DiffSyncModelFlags.IGNORE:
-            log.debug("Skipping due to IGNORE flag on source object")
-            return None
-        if dst_obj and dst_obj.model_flags & DiffSyncModelFlags.IGNORE:
-            log.debug("Skipping due to IGNORE flag on dest object")
-            return None
-
-        diff_element = DiffElement(
-            obj_type=model,
-            name=shortname,
-            keys=keys,
-            source_name=self.src_diffsync.name,
-            dest_name=self.dst_diffsync.name,
-            diff_class=self.diff_class,
-        )
-
-        if src_obj:
-            diff_element.add_attrs(source=src_obj.get_attrs(), dest=None)
-        if dst_obj:
-            diff_element.add_attrs(source=None, dest=dst_obj.get_attrs())
-
-        # Recursively diff the children of src_obj and dst_obj and attach the resulting diffs to the diff_element
-        self.diff_child_objects(diff_element, src_obj, dst_obj)
-
-        return diff_element
-
-    def diff_child_objects(
-        self, diff_element: DiffElement, src_obj: Optional[DiffSyncModel], dst_obj: Optional[DiffSyncModel],
-    ):
-        """For all children of the given DiffSyncModel pair, diff recursively, adding diffs to the given diff_element.
-
-        Helper method to `calculate_diffs`, usually doesn't need to be called directly.
-
-        These helper methods work in a recursive cycle:
-        diff_object_list -> diff_object_pair -> diff_child_objects -> diff_object_list -> etc.
-        """
-        children_mapping: Mapping[str, str]
-        if src_obj and dst_obj:
-            # Get the subset of child types common to both src_obj and dst_obj
-            src_mapping = src_obj.get_children_mapping()
-            dst_mapping = dst_obj.get_children_mapping()
-            children_mapping = {}
-            for child_type, child_fieldname in src_mapping.items():
-                if child_type in dst_mapping:
-                    children_mapping[child_type] = child_fieldname
-        elif src_obj:
-            children_mapping = src_obj.get_children_mapping()
-        elif dst_obj:
-            children_mapping = dst_obj.get_children_mapping()
-        else:
-            raise RuntimeError("Called with neither src_obj nor dest_obj??")
-
-        for child_type, child_fieldname in children_mapping.items():
-            # for example, child_type == "device" and child_fieldname == "devices"
-
-            # for example, getattr(src_obj, "devices") --> list of device uids
-            #          --> src_diffsync.get_by_uids(<list of device uids>, "device") --> list of device instances
-            src_objs = self.src_diffsync.get_by_uids(getattr(src_obj, child_fieldname), child_type) if src_obj else []
-            dst_objs = self.dst_diffsync.get_by_uids(getattr(dst_obj, child_fieldname), child_type) if dst_obj else []
-
-            for child_diff_element in self.diff_object_list(src=src_objs, dst=dst_objs):
-                diff_element.add_child(child_diff_element)
-
-        return diff_element
